// lookup = (name) => returns <config>
const { v4: uuidv4 } = require('uuid')
// const { uuid: uuidv4 } = require('uuidv4')
const client = require('../client')
const _ = require('lodash')

const addWord = (baseConfig, config) => ({ word, id, initial }) => {
  if (!baseConfig.config.words) {
    baseConfig.config.words = []
  }
  const words = baseConfig.config.words
  const def = { id, initial, uuid: config.uuid }
  if (words[word]) {
    words[word].push(def)
  } else {
    words[word] = [def]
  }
}

function isLetter (char) {
  return (/[a-zA-Z]/).test(char)
}

function configDup (config) {
  if (config instanceof Config) {
    return config.copy()
  }
  return _.cloneDeep(config)
}

function setWordsUUIDs (words, uuid) {
  for (const key in words) {
    words[key] = words[key].map((o) => Object.assign(o, { uuid }))
  }
}

function applyUUID (config, uuid) {
  if (config instanceof Config) {
    config = config.config
  }

  if (config.namespaces) {
    keys = Object.keys(config.namespaces)
    if (keys.length > 1) {
      debugBreak()
      // throw 'wtf 23'
    }
    if (config.namespaces[keys[0]]) {
      config.namespaces[uuid] = config.namespaces[keys[0]]
    }
  }
  if (config.operators) {
    config.operators = config.operators.map((o) => Object.assign(o, { uuid }))
  }
  if (config.bridges) {
    config.bridges = config.bridges.map((o) => Object.assign(o, { uuid }))
  }
  if (config.words) {
    setWordsUUIDs(config.words, uuid)
  }
  if (config.generators) {
    config.generators.forEach((generator) => generator[2] = uuid)
  }
  if (config.semantics) {
    config.semantics.forEach((semantic) => semantic[2] = uuid)
  }
}

function namespace_valid (namespace) {
  if (Array.isArray(namespace)) {
    return true
  }
  if (!namespace) {
    return true
  }
  return false
}

debugBreak = () => {
  debugger
}

class KM {
  toNS (id) {
    if (this._namespace.length == 0) {
      return id
    }
    return `${this._namespace.join('#')}#${id}`
  }

  getName (config) {
    if (config instanceof Config) {
      return config.config.name
    } else {
      return config.name
    }
  }

  get name() {
    return this.toNS(this._name)
  }

  constructor ({ config, namespace = [], uuid, isSelf = false }) {
    if (uuid) {
      this._uuid = uuid
      this._class = 'KM'
      this._config = config
      this._namespace = namespace
      this._name = this.getName(config)
      this._isSelf = true // new version the entry is for the containiner class but can't be that due to circularity
    } else {
      // this._uuid = uuidv4();
      if (config instanceof Config) {
        config.valid()
        this._name = config.config.name
        // config.uuid = this._uuid
        this._uuid = config._uuid
      } else {
        this._uuid = uuidv4()
        this._name = config.name
      }
      this._class = 'KM'
      this._config = config
      // applyUUID(this._config, this._uuid)
      if (this._config instanceof Config) {
        this._config.valid()
      }
      this._namespace = namespace
      this._isSelf = false // old version the entry is for the containiner class but can't be that due to circularity
    }
  }

  valid () {
    if (this._config.initializerFn && !(this._config instanceof Config)) {
      debugBreak()
      return false
    }
    if (this._namespace && !Array.isArray(this._namespace)) {
      debugBreak()
      return false
    }
    if (this._config instanceof Config) {
      if (!this._config.valid()) {
        debugBreak()
        return false
      }
    }
    return true
  }

  copy2 () {
    const config = configDup(this._config)
    const km = new KM({
      config,
      name: this._name,
      _uuid: config._uuid,
      namespace: this._namespace,
      isSelf: this._isSelf
    })
    return km // copy2()
  }

  copy () {
    const km = new KM({
      name: this._name,
      config: configDup(this._config),
      _uuid: uuidv4(),
      namespace: this._namespace,
      isSelf: this._isSelf
    })
    if (km.config instanceof Config) {
      km.config.valid()
    }
    applyUUID(km._config, km._uuid)
    return km // copy()
  }

  toString () {
    return `KM(${this._name}, ${JSON.stringify(this.config)}, ${this._namespace} isSelf(${this._isSelf}))`
  }

  get uuid () {
    return this._uuid
  }

  get config () {
    return this._config
  }

  set config (config) {
    this._config = config
  }

  get isSelf () {
    return this._isSelf
  }

  get namespace () {
    return this._namespace
  }

  set namespace (namespace) {
    this._namespace = namespace
  }
}

const multiApiImpl = (initializer) => {
  return {
    multiApi: true,

    // multi functions
    add: (config, multiApi, api) => {
      initializer(config, api)
      const name = api.getName()
      multiApi.apis[name] = api
      api.objects = config.get('objects')
      multiApi.current = name
    },

    initialize: ({ config, api: multiApi }) => {
      for (apiName in multiApi.apis) {
        const api = multiApi.apis[apiName]
        initializer(config, api)
      }
    },

    set objects (value) {
      for (key in Object.keys(this.apis)) {
        this.apis[key].objects = value
      }
    },

    // "product1": apiInstance(testData1),
    apis: {
    },

    // api functions
    api: (multiApi) => multiApi.apis[multiApi.current]
  }
}

class Config {
  defaultConfig () {
    this.config = {
      operators: [], // TODO
      bridges: [], // Done
      hierarchy: [], // Done
      name: '',
      namespaces: {
        // "config id" : { namespace: [""], ids: set(ids) }
      },
      eqClasses: [],
      priorities: [], // Done
      version: '3',
      debug: false,
      associations: { // Done
        negative: [],
        positive: []
      },
      objects: {
        // this is where the namespaced configs have their objects
        namespaced: {}
      },
      description: '',
      words: {}, // Done
      floaters: [],
      implicits: [],
      flatten: [],
      contexts: [], // TODO future
      generators: [], // Done
      semantics: [], // Done

      debug: false,
      expected_generated: [],
      expected_results: [],
      skipSemantics: false
    }
    // this.wasInitialized = false
  }

  objects () {
    return this.config.objects.namespaced[this._uuid]
  }

  addBridge (bridge) {
    if (!this.config.bridges) {
      this.config.bridges = []
    }
    const bridges = this.config.bridges
    const def = Object.assign({}, bridge, { uuid: this._uuid })
    bridges.push(def)
  }

  addGenerator (matcher, action) {
    if (!(typeof matcher === 'function')) {
      throw 'addGenerator: Expected matcher to be a function'
    }
    if (!(typeof action === 'function')) {
      throw 'addGenerator: Expected action to be a function'
    }

    if (!this.config.generators) {
      this.config.generators = []
    }
    const generators = this.config.generators
    const def = [matcher, action, this._uuid]
    generators.unshift(def)
  }

  addSemantic (matcher, action) {
    if (!(typeof matcher === 'function')) {
      throw 'addSemantic: Expected matcher to be a function'
    }
    if (!(typeof action === 'function')) {
      throw 'addSemantic: Expected action to be a function'
    }

    if (!this.config.semantics) {
      this.config.semantics = []
    }
    const semantics = this.config.semantics
    const def = [matcher, action, this._uuid]
    semantics.unshift(def)
  }

  addOperator (pattern) {
    if (!this.config.operators) {
      this.config.operators = []
    }
    const operators = this.config.operators
    const def = { pattern, uuid: this._uuid }
    operators.push(def)
  }

  addWord (word, def) {
    if (!this.config.words) {
      this.config.words = []
    }
    const words = this.config.words
    def = Object.assign({}, def, { uuid: this._uuid })
    if (words[word]) {
      words[word].push(def)
    } else {
      words[word] = [def]
    }
  }

  getAPI (uuid) {
    if (this._uuid == uuid) {
      return this.api
    }
    for (const km of this.configs) {
      if (km._uuid == uuid) {
        return km.config.api
      }
    }
  }

  server (server, key) {
    this._server = server
    this._key = key
  }

  getParams () {
    return {
      generators: this.config.generators,
      semantics: this.config.semantics,
      objects: this.config.objects,
      config: this
    }
  }

  processContexts (contexts, params) {
    client.processContexts(contexts, this.getParams())
  }

  processContext (context) {
    return client.processContext(context, this.getParams())
  }

  process (query) {
    return client.process(this, query)
  }

  initDefaults () {
    const init = (config) => {
      if (!config.objects) {
        config.objects = {
          namespaced: {
          }
        }
      }

      if (!config.associations) {
        config.associations = {}
      }

      if (!config.associations.positive) {
        config.associations.positive = []
      }

      if (!config.associations.negative) {
        config.associations.negative = []
      }
    }
    init(this.config)
    init(this.initConfig)
  }

  getConfig (name) {
    if (this.name == name) {
      return this
    }
    for (const config of this.configs) {
      if (config.config instanceof Config) {
        if (config.config.nsToString(config.config.name) == name) {
          return config.config
        }
      }
    }
  }

  // configs = [ { config, namespace } ... ]
  constructor (config) {
    if (config instanceof Config) {
      throw 'Excepted the config argument to be a hash not a Config object'
    }

    // set the default server so stuff just works
    this.server('http://184.67.27.82', '6804954f-e56d-471f-bbb8-08e3c54d9321')

    this.defaultConfig()
    this.initializerFn = ({ currentConfig }) => {
      if (currentConfig instanceof Config) {
        currentConfig.wasInitialized = true
      }
    }
    this.wasInitialized = false
    this.configs = []
    this._api = undefined
    this._namespace = []
    this._eqClasses = []
    this._uuid = uuidv4()
    if (config) {
      config = _.cloneDeep(config)
      this.config = config
      if (config.operators) {
        config.operators = config.operators.map((operator) => {
          if (typeof operator === 'string') {
            return { pattern: operator }
          } else {
            return operator
          }
        })
      }
    }
    this.initConfig = _.cloneDeep(this.config)
    this.configs.push(new KM({ config: this.config, uuid: this._uuid }))

    /*
    if (config) {
      this.configs.push(new KM({config, isSelf: true}))
      this.addInternal(Object.assign({}, config), false)
    } else {
      this.configs.push( new KM({config: this.config, isSelf: true}) )
    }
    */

    this.setUUIDs()
    this.initDefaults()
    if (!this.config.objects.namespaced) {
      this.config.objects.namespaced = {}
    }
    this.get('objects').namespaced[this._uuid] = {}
    this.valid()
  }

  set multiApi (initializer) {
    this.api = multiApiImpl(initializer)
  }

  get api () {
    if (this._api && this._api.multiApi) {
      return this._api.api(this._api)
    } else {
      return this._api
    }
  }

  set api (value) {
    if (this._api && this._api.multiApi) {
      this._api.add(this, this._api, value)
    } else {
      this._api = value
      if (this._api) {
        this._api.objects = this.config.objects
      }
    }
  }

  set afterTest (value) {
    this._afterTest = value
  }

  get uuid () {
    return this._uuid
  }

  set uuid (uuid) {
    const map = { [this._uuid]: uuid }
    this._uuid = uuid
    this.mapUUIDs(map)
  }

  // remove all added modules and initialize with the init config
  resetToOne () {
    /*
    this.config = _.cloneDeep(this.initConfig)
    this.configs = [this.configs[0]]
    this.setUUIDs()
    this.initDefaults()
    */
    this.configs = [this.configs[0]]
    this.defaultConfig()
    if (!this.initializeFn) {
      // TODO move this to the default initializer
      Object.assign(this.config.objects, _.cloneDeep(this.initConfig.objects || {}))
    }
    this.initializeConfigFromConfigs({ others: [], objects: this.config.objects.namespaced, moreNames: [] })
    const map = {}
    for (let i = 0; i < this.configs.length; ++i) {
      map[this.configs[i].uuid] = this.configs[i].uuid
    }
    this.mapUUIDs(map)
    this.valid()
  }

  copy () {
    this.valid()
    const cp = new Config()
    cp.configs = this.configs.map((km) => km.copy2())
    cp._uuid = cp.configs[0]._uuid
    cp.initializerFn = this.initializerFn
    cp._api = this._api
    cp._namespace = this._namespace
    cp._eqClasses = this._eqClasses
    cp.name = this.name
    cp.description = this.description
    cp.tests = this.tests

    cp.initConfig = _.cloneDeep(this.initConfig)

    cp.defaultConfig()
    // cp.wasInitialized = false;  // since default config GREG
    cp.resetWasInitialized()
    if (!this.initializeFn) {
      // TODO move this to the default initializer
      Object.assign(cp.config.objects, _.cloneDeep(this.initConfig.objects || {}))
    }
    cp.initializeConfigFromConfigs({ others: [], objects: cp.config.objects.namespaced, moreNames: [] })
    const map = {}
    for (let i = 0; i < this.configs.length; ++i) {
      map[this.configs[i].uuid] = cp.configs[i].uuid
    }
    cp.mapUUIDs(map)
    cp.valid()
    return cp
  }

  setUUIDs () {
    this.config.bridges && this.config.bridges.forEach((bridge) => bridge.uuid = this._uuid)
    this.config.words && setWordsUUIDs(this.config.words, this._uuid)
    this.config.operators && this.config.operators.forEach((operator) => operator.uuid = this._uuid)
    const ids = (this.config.bridges && this.config.bridges.map((bridge) => bridge.id)) || []
    ids.sort()
    this.config.namespaces = {}
    if (true || ids.length > 0) {
      this.config.namespaces[this.uuid] = { ids, namespace: this._namespace }
    }
    if (this.config.generators) {
      this.config.generators.forEach((generator) => generator[2] = this._uuid)
    }
    if (this.config.semantics) {
      this.config.semantics.forEach((semantic) => semantic[2] = this._uuid)
    }
  }

  mapUUIDs (map) {
    if (this._eqClasses) {
      this._eqClasses = this._eqClasses.map((eqclass) => eqclass.map((uuid) => map[uuid] || uuid))
    }

    if (this.config.namespaces) {
      const ns = {}
      for (const uuid of Object.keys(this.config.namespaces)) {
        const uuidPrime = map[uuid] || uuid
        ns[uuidPrime] = this.config.namespaces[uuid]
      }
      this.config.namespaces = ns
    }

    if (this.config.words) {
      const words = this.config.words
      for (const key in words) {
        words[key].forEach((o) => {
          if (o.uuid) {
            if (map[o.uuid]) {
              o.uuid = map[o.uuid]
            }
          } else {
            o.uuid = this._uuid
          }
        })
      }
    }

    if (this.config.bridges) {
      this.config.bridges.forEach((bridge) => {
        if (bridge.uuid) {
          if (map[bridge.uuid]) {
            bridge.uuid = map[bridge.uuid]
          }
        } else {
          bridge.uuid = this._uuid
        }
      })
    }

    if (this.config.operators) {
      this.config.operators.forEach((operator) => {
        if (operator.uuid) {
          if (map[operator.uuid]) {
            operator.uuid = map[operator.uuid]
          }
        } else {
          operator.uuid = this._uuid
        }
      })
    }

    if (this.config.generators) {
      this.config.generators.forEach((generator) => {
        if (generator[2]) {
          if (map[generator[2]]) {
            generator[2] = map[generator[2]]
          }
        } else {
          generator[2] = this._uuid
        }
      })
    }

    if (this.config.semantics) {
      this.config.semantics.forEach((semantic) => {
        if (semantic[2]) {
          if (map[semantic[2]]) {
            semantic[2] = map[semantic[2]]
          }
        } else {
          semantic[2] = this._uuid
        }
      })
    }
  }

  initializeFromConfigs () {
    this.configs.forEach(({ config, namespace, uuid }) => {
      /*
      let objects = this.get('objects')
      if (namespace) {
        objects = {}
        this.get('objects')['namespaced'][namespace] = objects
      }
      */
      const objects = {}
      if (config instanceof Config) {
        const aw = addWord(this, config)
        this.get('objects').namespaced[config._uuid] = objects
        config.initializerFn({ addWord: aw, config, baseConfig: this, currentConfig: config, objects, namespace, uuid, api: config.api })
      } else {
        const aw = addWord(this, this)
        this.get('objects').namespaced[this._uuid] = objects
        this.initializerFn({ addWord: aw, config: this, baseConfig: this, currentConfig: this, objects, namespace, uuid, api: this.api })
      }
    })
  }

  initialize ({ force = true } = {}) {
    if (force || !this.wasInitialized) {
      const aw = addWord(this, this)
      this.initializerFn({ addWord: aw, config: this, baseConfig: this, currentConfig: this, objects: this.get('objects'), uuid: this._uuid, namespace: '', api: this.api })
      this.wasInitialized = true
    }
  }

  initializer (fn) {
    this.wasInitialized = false
    this.initializerFn = (args) => {
      const baseConfig = args.baseConfig
      const currentConfig = args.currentConfig
      if (!currentConfig.wasInitialized) {
        if (currentConfig.api) {
          currentConfig.api.objects = args.objects
        }
        fn(args)
        currentConfig.wasInitialized = true
      }
    }
    this.initializeFromConfigs()
  }

  nsToString (id) {
    return `${this._namespace.concat(id).join('#')}`
  }

  valid () {
    const uuids = this.configs.map((config) => config._uuid)

    if (this.config.generators) {
      const problem = this.config.generators.some((generator) => {
        if (generator.length != 3) {
          debugBreak()
          return true
        }
        if (!uuids.includes(generator[2])) {
          debugBreak()
          return true
        }
      })

      if (problem) {
        return false
      }
    }

    if (this.config.semantics) {
      const problem = this.config.semantics.some((semantic) => {
        if (semantic.length != 3) {
          debugBreak()
          return true
        }
        if (!uuids.includes(semantic[2])) {
          debugBreak()
          return true
        }
      })

      if (problem) {
        return false
      }
    }

    if (this._uuid != this.configs[0].uuid) {
      debugBreak()
      return false
    }
    if (this.configs[0]._config instanceof Config) {
      debugBreak()
      return false
    }

    if (!this.config.objects) {
      debugBreak()
      return false
    }

    for (const key in this.config.words) {
      const values = this.config.words[key]
      if (values.some((word) => (Object.keys(word).includes('uuid') && !word.uuid))) {
        debugBreak()
        return false
      }
    }

    const kms_uuids = this.configs.map((km) => km.uuid)
    const bridges_uuids = (this.config.bridges && this.config.bridges.map((bridge) => bridge.uuid).filter((uuid) => uuid)) || []
    let result = true
    bridges_uuids.forEach((buuid) => {
      if (!kms_uuids.includes(buuid)) {
        debugBreak()
        result = false
      }
    })

    if (this.get('namespaces')) {
      const namespace_uuids = Object.keys(this.get('namespaces'))
      namespace_uuids.forEach((uuid) => {
        if (!kms_uuids.includes(uuid)) {
          debugBreak() // in namespaces
          result = false
        }
      })
      if (!result) {
        debugBreak()
        return false
      }
      kms_uuids.forEach((uuid) => {
        if (!namespace_uuids.includes(uuid)) {
          debugBreak() // in namespaces
          result = false
        }
      })
      if (!result) {
        debugBreak()
        return false
      }
    }

    if (Object.keys(this.config.namespaces || {}) == ['undefined']) {
      debugBreak()
      result = false
    }
    if (!result) {
      debugBreak() // not valid
      return false
    }

    result = true
    this.configs.forEach((km) => {
      if (!km.valid()) {
        result = false
      }
    })
    if (!result) {
      debugBreak() // not valid
      return false
    }
    return result
  }

  resetWasInitialized () {
    this.wasInitialized = false
    this.configs.forEach((km) => {
      const config = km.config
      if (config instanceof Config) {
        config.wasInitialized = false
      }
    })
  }

  rebuild ({ isModule: mainIsModule = false } = {}) {
    this.config = _.cloneDeep(this.initConfig)
    this.config.objects.namespaced = {}
    this.resetWasInitialized()
    this.configs.forEach((km) => {
      const namespace = km.namespace
      // const namespaced_objects = this.path_to_namespace(this.config.objects.namespaced, namespace)
      this.config.objects.namespaced[km._uuid] = {}
      const namespaced_objects = this.config.objects.namespaced[km._uuid]
      this.setupNamespace(km)
      const aw = addWord(km.config, km.config)
      let config = km.config
      let isSelf = false
      let isModule = true
      if (!(config instanceof Config)) {
        config = this
        isSelf = true
        isModule = mainIsModule
      }
      if (!isSelf) {
        config.config = _.cloneDeep(config.initConfig)
      }
      config.wasInitialized = false
      // TODO change name of config: to baseConfig:
      config.initializerFn({ isModule, addWord: aw, config, baseConfig: this, currentConfig: config, uuid: config._uuid, objects: namespaced_objects, namespace, api: config.api })
      if (config._api && config._api.initialize) {
        config._api.initialize({ config, api: config._api })
      }
      config.setUUIDs()
      config.applyNamespace(config.config, namespace, config.uuid)
      if (!isSelf) {
        this.addInternal(config, true, false, false, true)
      }
      km.valid()
    })
    this.valid()
  }

  // name: namespace name
  // others
  //    if undefined namespace applies to first loaded config
  //    if list of name then namespace applied only to those configs
  //    if true then namespace applies to all loaded configs
  namespace (moreNames, others) {
    const config = this.copy()
    if (!Array.isArray(moreNames)) {
      moreNames = [moreNames]
    }
    config.defaultConfig()

    if (Array.isArray(others)) {
      config.configs.forEach((km) => {
        let config = km.config
        let updateNamespace = false
        if (!(km.config instanceof Config)) {
          config = this
          updateNamespace = true
        }
        if (others.includes(config.config.name)) {
          km.namespace = moreNames.concat(km.namespace)
          if (updateNamespace) {
            config._namespace = moreNames.concat(config._namespace)
          }
        }
      })
    } else if (others) {
      config.configs.forEach((km) => {
        km.namespace = moreNames.concat(km.namespace)
      })
      config._namespace = moreNames.concat(config._namespace)
    } else {
      const km = config.configs[0]
      km.namespace = moreNames.concat(km.namespace)
      config._namespace = moreNames.concat(config._namespace)
    }
    config.rebuild()
    return config
  }

  setupNamespace (km) {
    let config = km.config
    if (!(config instanceof Config)) {
      config = this.initConfig
    }
    const bridges = ((config instanceof Config) ? config.config.bridges : config.bridges) || []
    const ids = bridges.map((bridge) => bridge.id)
    this.config.namespaces = this.config.namespaces || {}

    this.config.namespaces[km.uuid] = {
      ids,
      namespace: km.namespace
    }
  }

  // config.initializerFn({ config: this, objects, namespace })
  // this goes through all the configs and sets up this.config
  initializeConfigFromConfigs ({ others, objects, moreNames }) {
    // setup the namespace in the configs
    let first = true
    this.configs.forEach((km) => {
      let config = km.config
      if (!(config instanceof Config)) {
        config = this.initConfig
      }
      if (!km.namespace) {
        km.namespace = []
      }
      if (config instanceof Config) {
        //km.namespace = config._namespace
        config = config.initConfig
      }

      const configPrime = _.cloneDeep(config)

      if (Array.isArray(others)) {
        if (others.indexOf(config.name) >= 0) {
          km.namespace = moreNames.concat(km.namespace)
        }
      } else if (others === true || (!others && first)) {
        km.namespace = moreNames.concat(km.namespace)
      }

      this.applyNamespace(configPrime, km.namespace, km.uuid)
      first = false
      this.addInternal(configPrime, false, true)
      applyUUID(configPrime, km.uuid)
    })

    // setup the namespaces mapping for passing to the server side

    this.configs.forEach((km) => {
      const namespace = km.namespace
      if (true || namespace.length > 0) { this.setupNamespace(km) }
    })

    // initialize the configs in context

    this.configs.forEach(({ namespace, uuid, config, isSelf }) => {
      if (isSelf) {
        config = this
      }
      let nsobjects = objects
      if (namespace) {
        nsobjects = this.path_to_namespace(objects, namespace)
      }
      if (config instanceof Config) {
        const aw = addWord(this, config)
        config.initializerFn({ addWord: aw, baseConfig: this, currentConfig: config, config, objects: nsobjects, namespace, uuid, api: config.api })
      } else {
        const aw = addWord(this, this)
        this.initializerFn({ addWord: aw, baseConfig: this, currentConfig: this, config: this, objects: nsobjects, namespace, uuid, api: this.api })
      }
    })
  }

  applyNamespace (config, namespace, uuid) {
    const toNS = (id) => {
      if (namespace.length == 0) {
        return id
      }
      return `${namespace.join('#')}#${id}`
    }
    /*
    if (config.name) {
      config.name = toNS(config.name)
    }
    */
    if (config.associations) {
      const map = (associations) => {
        return associations.map((association) => {
          return association.map((id) => [toNS(id[0]), id[1]])
        })
      }
      const associations = config.associations
      if (associations.positive) {
        associations.positive = map(associations.positive)
      }
      if (associations.negative) {
        associations.negative = map(associations.negative)
      }
    }

    if (config.words) {
      const words = {}
      for (const word in config.words) {
        words[word] = config.words[word].map((word) => {
          return Object.assign({}, word, { id: toNS(word.id), uuid })
        })
      }
      config.words = words
    }

    if (config.hierarchy) {
      let hierarchy = config.hierarchy
      hierarchy = hierarchy.map((h) => {
        return h.map((id) => toNS(id))
      })
      config.hierarchy = hierarchy
    }

    if (config.priorities) {
      let priorities = config.priorities
      priorities = priorities.map((p) => {
        return p.map((id) => {
          return [toNS(id[0]), id[1]]
        })
      })
      config.priorities = priorities
    }

    if (config.generators) {
      let generators = config.generators
      generators = generators.map((g) => {
        return [g[0], g[1], uuid]
      })
      config.generators = generators
    }

    if (config.semantics) {
      let semantics = config.semantics
      semantics = semantics.map((s) => {
        return [s[0], s[1], uuid]
      })
      config.semantics = semantics
    }
  }

  checkBridges (bridges) {
    const keys = {}
    bridges.forEach((bridge) => {
      const key = JSON.stringify({ id: bridge.id, level: bridge.level })
      if (keys[key]) {
        throw `The operator ${bridge.id}/${bridge.level} appears more than once in the bridges.`
      } else {
        keys[key] = true
      }
    })
  };

  get (property) {
    return this.config[property]
  }

  set (property, value) {
    if (!this.config.hasOwnProperty(property)) {
      throw `Setting invalid property ${property}`
    }
    this.config[property] = value
    //this.configs[0][property] = value
  }

  path_to_namespace (objects, path) {
    let ns = objects
    path = Object.assign([], path)
    while (path.length > 0) {
      const next = path.shift()
      if (!ns[next]) {
        ns[next] = {}
      }
      ns = ns[next]
    }
    return ns
  }

  addEqClass(existing, other) {
    for (const eqClass of this._eqClasses) {
      if (eqClass.includes(other._uuid)) {
        eqClass.push(other._uuid)
      }
    }
  }

  add (more) {
    if (more === this) {
      throw 'Cannot add an object to itself.'
    }
    if (!(more instanceof Config)) {
      more = new Config(more)
    }

    this.valid()
    more.valid()
    // copy so i don't have to copy later
    more = more.copy()

    // get the new ones
    // remove the dups
    // run the initialize one all new ones
    // this.configs = this.configs.concat(new KM({config: more}));
    // only set for the first one the rest have it set
    const namespace = this._namespace.concat(more._namespace)
    const moreKMs1 = [new KM({ config: more, uuid: more.uuid, namespace })]
    const moreKMs2 = more.configs.slice(1).map((km) => {
      return km
      //const cp = km.copy()
      //cp.namespace = namespace
      //return cp;
    })
    const moreKMs = moreKMs1.concat(moreKMs2)
    const eqClass = moreKMs.map((km) => km.uuid)
    this._eqClasses.push(eqClass)
    // look for dups and combine them with the eqclasses
    for (const moreKM of moreKMs) {
      const existingKM = this.configs.find( (km) => moreKM._name && km.name == moreKM.name )
      if (existingKM) {
        this.addEqClass(existingKM, moreKM)
      } else {
        this.configs.push(moreKM)
      }
    }
    more.resetToOne()
    this.config.eqClasses = this._eqClasses
    this.rebuild()
    this.valid()
    return this
  }

  // TODO get rid of useOldVersion arg
  addInternal (more, useOldVersion = true, skipObjects = false, includeNamespaces = true, allowNameToBeNull = false) {
    if (more instanceof Config) {
      more.initialize({ force: false })
      if (useOldVersion) {
        more = more.config
      } else {
        // more = more.initConfig
        more = _.cloneDeep(more.initConfig)
      }
    }
    for (const key of Object.keys(more)) {
      const value = more[key]
      // TODO remove name and description on the config bag
      const noOverwrite = ['name', 'namespaced']
      if (!this.config[key]) {
        if (allowNameToBeNull) {
          if (noOverwrite.includes(key)) {
            continue
          }
        } else if (this.config[key] && noOverwrite.includes(key)) {
          continue
        }
        this.config[key] = value
        continue
      }
      if (key == 'words') {
        const config_words = this.config.words
        const more_words = more.words
        for (const word of Object.keys(more_words)) {
          if (!config_words[word]) {
            config_words[word] = []
          }
          config_words[word] = config_words[word].concat(more_words[word])
        }
      } else if (key == 'name') {
        /*
        if (this.config[key] == '') {
          this.config[key] = more[key]
        }
        */
      } else if (key == 'namespaces') {
        if (includeNamespaces) {
          Object.assign(this.config[key], more[key])
        }
      } else if (key == 'debug') {
        this.config[key] = this.config[key] || more[key]
      } else if (key == 'description') {
        this.config[key] += ' ' + more[key].trim()
      } else if (key == 'objects') {
        if (!skipObjects) {
          // namespaced is special
          const namespaced = this.config.objects.namespaced
          Object.assign(this.config[key], more[key])
          this.config.objects.namespaced = namespaced
        }
      } else if (key == 'associations') {
        const config_associations = this.config.associations
        const more_associations = more.associations
        if (more_associations.positive) {
          config_associations.positive = config_associations.positive.concat(more_associations.positive)
        }
        if (more_associations.negative) {
          config_associations.negative = config_associations.negative.concat(more_associations.negative)
        }
      } else if (Array.isArray(value)) {
        // console.log('key', key, 'XXX')
        // console.log('more', JSON.stringify(more, null, 2))
        this.config[key] = this.config[key].concat(more[key])
      } else {
        if (!(key in this.config)) {
          throw `Unexpected property in config ${key}`
        }
        this.config[key] = more[key]
      }
    }
    return this
  }
}

module.exports = {
  Config
}
