const { args } = require('./helpers')

class Semantic {
  constructor (matcher, action, uuid) {
    this.matcher = matcher
    this.action = action
    this.uuid = uuid
  }

  toString () {
    return `Semantic(${this.matcher}, ${this.action})`
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  matches (objects, context, hierarchy, config) {
    // throwing for a match is false
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    return this.matcher({ args: args(context, hierarchy), objects, global: objects, context, hierarchy, api: this.getAPI(config) })
  }

  apply (objects, context, hierarchy, config, response, s) {
    const context_prime = Object.assign({}, context)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    this.action({ args: args(context, hierarchy), objects, global: objects, n, hierarchy, context: context_prime, uuid: this.uuid, config, response, s, api: this.getAPI(config) })
    return context_prime
  }
}

class Semantics {
  constructor (semantics) {
    this.semantics = semantics
    this.logs = []
  };

  applyToContext (objects, context, hierarchy, config, response) {
    let context_prime = {}
    if (!(context instanceof Array || context instanceof Object)) {
      return context
    }

    for (const key in context) {
      context_prime[key] = this.apply(objects, context[key], hierarchy, config, response)
    }
    const s = (c) => this.applyToContext(objects, c, hierarchy, config, response)
    for (const isemantic in this.semantics) {
      const semantic = this.semantics[isemantic]
      if (semantic.matches(objects, context, hierarchy, config)) {
        context_prime = semantic.apply(objects, context, hierarchy, config, response, s)
        this.logs.push(`Semantics: applied ${semantic.toString()}\n to\n ${JSON.stringify(context)}\n the result was ${JSON.stringify(context_prime)}\n`)
        break
      }
    }
    return context_prime
  }

  applyToContexts (objects, contexts, hierarchy, config) {
    const contexts_prime = []
    contexts.forEach((context) => {
      contexts_prime.push(this.applyToContext(objects, context, hierarchy, config))
    })
    return contexts_prime
  }

  apply (objects, value, hierarchy, config, response) {
    if (Array.isArray(value)) {
      return this.applyToContexts(objects, value, hierarchy, config, response)
    } else if (value instanceof Object) {
      return this.applyToContext(objects, value, hierarchy, config, response)
    } else {
      return value
    }
  }
}

module.exports = { Semantic, Semantics }
